#include "VolumeRenderer.h"
#include "global.h"
#include <QtCore\qmath.h>

VolumeRenderer::VolumeRenderer(Volume* volume) :
	volume(volume)
{
	loadUpQuadVAO();
	initFramebuffers();

	dof = new DoFShader();
	quad = new QuadShader();
    tex3D = new Tex3DShader();

    glGenVertexArrays(1, &vao);
    glGenBuffers(1, &vboVertices);
    glGenBuffers(1, &vboTextureCoord);

    glBindVertexArray(vao);
    glBindBuffer(GL_ARRAY_BUFFER, vboVertices);
    glBufferData(GL_ARRAY_BUFFER, 6 * sizeof(QVector3D), NULL, GL_STREAM_DRAW);
    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, 0);

    glBindBuffer(GL_ARRAY_BUFFER, vboTextureCoord);
    glBufferData(GL_ARRAY_BUFFER, 6 * sizeof(QVector3D), NULL, GL_STREAM_DRAW);
    glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, 0);

    glBindVertexArray(0);
    glBindBuffer(GL_ARRAY_BUFFER, 0);

	focusPlane = 0;
	blurAccum = 45/2.f;
	blurAttenuation = qTan(qDegreesToRadians(blurAccum));
}

VolumeRenderer::~VolumeRenderer() {
	glDeleteVertexArrays(1, &vao);
	glDeleteBuffers(1, &vboTextureCoord);
	glDeleteBuffers(1, &vboVertices);
	glDeleteTextures(1, &texID);

	for (int i = 0; i < fbos.size(); i++)
	{
		delete fbos[i];
	}
	fbos.clear();

	delete dof;
	dof = nullptr;

	delete quad;
	quad = nullptr;

    delete proxyGeometry;
    proxyGeometry = nullptr;
}

void VolumeRenderer::loadUpQuadVAO() {
    std::vector<QVector2D> vertices;
    vertices.push_back(QVector2D(-1, -1));
    vertices.push_back(QVector2D(1, -1));
    vertices.push_back(QVector2D(-1, 1));
    vertices.push_back(QVector2D(1, 1));

    std::vector<QVector2D> textCoord;
    textCoord.push_back(QVector2D(0, 0));
    textCoord.push_back(QVector2D(1, 0));
    textCoord.push_back(QVector2D(0, 1));
    textCoord.push_back(QVector2D(1, 1));

    glGenVertexArrays(1, &vaoQuad);
    glGenBuffers(1, &vboQuadVertices);
    glGenBuffers(1, &vboQuadTextureCoord);

    glBindVertexArray(vaoQuad);
    glBindBuffer(GL_ARRAY_BUFFER, vboQuadVertices);
    glBufferData(GL_ARRAY_BUFFER, vertices.size() * sizeof(QVector2D), &vertices[0], GL_STATIC_DRAW);
    glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, 0, NULL);

    glBindBuffer(GL_ARRAY_BUFFER, vboQuadTextureCoord);
    glBufferData(GL_ARRAY_BUFFER, textCoord.size() * sizeof(QVector2D), &textCoord[0], GL_STATIC_DRAW);
    glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 0, NULL);

    glBindVertexArray(0);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
}

void VolumeRenderer::loadUpVAO(std::vector<QVector3D>& vertices, std::vector<QVector3D>& texCoord) {

    glNamedBufferSubData(vboVertices, 0, vertices.size() * sizeof(QVector3D), &vertices[0]);
    glNamedBufferSubData(vboTextureCoord, 0, texCoord.size() * sizeof(QVector3D), &texCoord[0]);

}

void VolumeRenderer::loadUp3DTexture() {
	//convert voxel data
	const std::vector<Voxel>& voxels = volume->getVoxels();
	unsigned short bytesPerTexel = 1;
	std::vector<byte> texels;
	texels.resize(volume->size() * bytesPerTexel);

	for (unsigned int x = 0; x < volume->size(); x += 1) {
		texels[x] = (byte)(voxels[x].getValue() * 255.0f);
	}

	//upload 3d Texture
	glGenTextures(1, &texID);
	glBindTexture(GL_TEXTURE_3D, texID);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_BORDER);
	glTexImage3D(GL_TEXTURE_3D, 0, GL_R8, volume->width(), volume->height(), volume->depth(), 0, GL_RED, GL_UNSIGNED_BYTE, &texels[0]);

	glBindTexture(GL_TEXTURE_3D, 0);
}

void VolumeRenderer::initFramebuffers() {

	for (int i = 0; i < 5; i++)	{
		fbos.push_back(new FBO());
	}

}

void VolumeRenderer::clearFBO(const int& fbo)
{
	glBindFramebuffer(GL_FRAMEBUFFER, fbos[fbo]->getID());
	glViewport(0, 0, 1024, 1024);
	glClearColor(0, 0, 0, 0);
	glClear(GL_COLOR_BUFFER_BIT);
	glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

void VolumeRenderer::render(GLCamera& camera) {
	if (dof->isRenderable()) {
        proxyGeometry->updateProxyGeometry(camera.getModelMatrix().transposed(), numSamples);
        std::vector<std::vector<QVector3D>> polygons = proxyGeometry->getPlaneVertices();
        std::vector<std::vector<QVector3D>> texCoords = proxyGeometry->getTexCoords();

		float depth_distance = ((float)volume->depth()) / volume->width();
        float slice_distance = depth_distance / volume->depth();
        float sliceDistance = proxyGeometry->getSliceDistance();
		int prev = 0;
		int next = 1;

		clearFBO(prev);
		clearFBO(next);

		//B2F
		for (int s = 0; s <= focusPlane; s++) {

            clearFBO(4);
            tex3D->prepare(camera, vao, texID);
            loadUpVAO(polygons[numSamples-s-1], texCoords[numSamples-s-1]);
            tex3D->draw(fbos[4]);
            
            float depthCoord = s*sliceDistance;
            float C = slice_distance * qAbs(s - focusPlane) * blurAttenuation;

            dof->prepare(camera, vaoQuad, fbos[4]->getTXO());
            dof->loadDepth(depthCoord, sliceDistance);
            dof->draw(fbos[prev], fbos[next], C, false); //B2F

			int tmp = prev;
			prev = next;
			next = tmp;
			clearFBO(next);
			lastB2F = prev;
		}

		prev = 2;
		next = 3;
		clearFBO(prev);
		clearFBO(next);

        //F2B
        for (int s = numSamples; s > focusPlane; s--)
        {
            clearFBO(4);
            tex3D->prepare(camera, vao, texID);
            loadUpVAO(polygons[numSamples - s], texCoords[numSamples - s]);
            tex3D->draw(fbos[4]);

            float depthCoord = s*sliceDistance;
            float C = slice_distance * qAbs(s - focusPlane) * blurAttenuation;

            dof->prepare(camera, vaoQuad, fbos[4]->getTXO());
            dof->loadDepth(depthCoord, sliceDistance);
            dof->draw(fbos[prev], fbos[next], C, true); //F2B

			int tmp = prev;
			prev = next;
			next = tmp;
			clearFBO(next);
			lastF2B = prev;
		}

		glBindFramebuffer(GL_FRAMEBUFFER, 0);
		dof->cleanUp();
	}

}

void VolumeRenderer::render(int w, int h, GLCamera& camera) {
	if (quad->isRenderable()) {
		glViewport(0, 0, w, h);
		glClearColor(0, 0, 0, 1);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

		quad->prepare(vaoQuad);
		quad->draw(fbos[lastB2F]->getTXO(), camera);
		quad->draw(fbos[lastF2B]->getTXO(), camera);
		quad->cleanUp();

		glDisable(GL_BLEND);
		glDisable(GL_DEPTH_TEST);
	}

}

void VolumeRenderer::prepareNewVolume(Volume* v) {
    numSamples = 200;
	volume = v;
	loadUp3DTexture();
    QVector3D sizes = QVector3D(volume->width(), volume->height(), volume->depth());
    float max = qMax(sizes.x(), qMax(sizes.y(), sizes.z()));
    sizes = 2.0f * sizes / max;
    delete proxyGeometry;
    proxyGeometry = new ProxyGeometry(sizes);
	focusPlane = numSamples / 2;
	lastFocusPlane = numSamples - 1;
}

void VolumeRenderer::refreshShader()
{
	dof->refresh();
}

void VolumeRenderer::setFocusPlane(const int & value)
{
	if (volume != nullptr) {
		int tmp = focusPlane + value;
        if (0 <= tmp && tmp < numSamples) {
            focusPlane = tmp;
        }
	}
	else
		std::cout << "NO VOLUME DATA FOUND" << std::endl;

}

void VolumeRenderer::setBlurFactor(const int & value)
{
	int tmp = blurAccum + value;
	if (0 <= tmp && tmp <= 45) {
		blurAccum = tmp;
		blurAttenuation = qTan(qDegreesToRadians(blurAccum));
	}
}
