#pragma once

#include <glew.h>
#include <vector>
#include <QtGui\qopenglshaderprogram.h>

#include "GLCamera.h"
#include "GLShader.h"
#include "FBO.h"
#include "GUI\Volume.h"
#include "SHADER\DoFShader.h"
#include "SHADER\QuadShader.h"
#include "SHADER/Tex3DShader.h"
#include "RENDER/ProxyGeometry.h"

class VolumeRenderer {

    public:
        VolumeRenderer(Volume* volume);
        ~VolumeRenderer();

        /** prepare the new volume
         * loads up the 3d texture and sets the proxygeometry for the volume
         * @param v: the new volume
         */
        void prepareNewVolume(Volume* v);

        /**
         * renders the volume
         * @param camera: the camera containing the necessary transformation, view, projection-matrices
         */
		void render(GLCamera& camera);

        // TEST 2.0
        /** recompiles the depth of field shader
         * for debugging 
         */
        void refreshShader();
        /** changes the plane in focus
         * @param value: index of plane in focus
         */
        void setFocusPlane(const int& value);
        
        /** changes the effectiveness of the DoF effect
         * @param value: blur factor, the higher the value the greater the blur
         */
        void setBlurFactor(const int& value);

        /** blends F2B image and B2F image into the defalt framebuffer */
		void render(int w, int h, GLCamera& camera);

    private:
        /**
         * loads the 3D texture of the volume
         */
        void loadUp3DTexture();

        /**
         * loads the VAO for the quad
         */
        void loadUpQuadVAO();

        /**
         * updates the vertices and texture coordinates for a slice
         * @param vertices: list of new vertices
         * @param texCoord: list of new texture coordinates
         */
        void loadUpVAO(std::vector<QVector3D>& vertices, std::vector<QVector3D>& texCoord);

        /**
         * initializes the necessary framebuffers
         */
        void initFramebuffers();

        /**
         * clears the FBOs
         * @param fbo: index of fbo
         */
		void clearFBO(const int& fbo);

        /** volume to be rendered */
        Volume* volume = nullptr;
        /** Shader for the depth of field effect */
		DoFShader* dof = nullptr;
        /** Shader for rendering a simple quad */
		QuadShader* quad = nullptr;
        /** Shader for rendering the slices from a 3D texture */
        Tex3DShader* tex3D = nullptr;

        /** Proxygeometry holding the slice informations
         * e.g. intersection points, texture coordinates
         */
        ProxyGeometry* proxyGeometry = nullptr;

        /** list of Framebuffer objects 
         * 2x2 + 1 FBO: 2 for F2B, 2 for B2F, 1 for rendering a slice from the 3D texture
         */
		std::vector<FBO*> fbos;

        /** holds the index of the last B2F fbo */
        int lastB2F = -1;
        /** holds the index of the last F2B fbo */
        int lastF2B = -1;
        /** the number of slices that are drawn */
        int numSamples;

        /** texture ID of the 3D texture from the volume */
		GLuint texID;

        /** VAO for the slices of the proxy geometry */
        GLuint vao;
        /** VBO for the vertices from proxy geometry */
        GLuint vboVertices;
        /** VBO for the texture coordinates from proxy geometry */
        GLuint vboTextureCoord;

        /** VAO for a simple quad */
        GLuint vaoQuad;
        /** VBO for the quad vertices */
        GLuint vboQuadVertices;
        /** VBO for the quad texture coordinates */
        GLuint vboQuadTextureCoord;
		
};